const BookFusion = {}

BookFusion.Common = (() => {
  window.addEventListener('load', () => {
    // Notifies native code that the page is loaded after it is rendered.
    // Waiting for the next animation frame seems to do the trick to make sure the page is fully rendered.
    window.requestAnimationFrame(() => {
    // TODO test only
//        setTimeout(() => {
            BookFusionAppBridge.onDocumentLoaded()
//        }, 1000)
    })
  })

  // For setting user setting.
  function setProperty (key, value) {
    const root = document.documentElement
    root.style.setProperty(key, value)
  }

  // For removing user setting.
  function removeProperty (key) {
    const root = document.documentElement
    root.style.removeProperty(key)
  }

  // For setting multiple user setting.
  function setProperties (properties) {
    Object.entries(properties).forEach(([key, value]) => {
      if (value === null) {
        removeProperty(key)
      } else {
        setProperty(key, value)
      }
    })
  }

  function getProperty (key) {
    return document.documentElement.style.getPropertyValue(key).toString().trim()
  }

  function isScrollModeEnabled () {
    return getProperty('--BF__scrollMode') === 'bookfusion-scroll-on'
  }

  function calculateContentSize () {
    const pixelRatio = window.devicePixelRatio;

    if (isScrollModeEnabled()) {
      return {
        height: document.documentElement.getBoundingClientRect().height * pixelRatio,
        width: document.documentElement.clientWidth * pixelRatio
      }
    } else {
      const width = Math.max(
        document.documentElement.scrollWidth,
        document.body && document.body.scrollWidth // body can be undefined for parsererror
      )

      return {
        height: document.documentElement.clientHeight * pixelRatio,
        width: width * pixelRatio
      }
    }
  }

  let clientContentSize
  function getContentSize () {
    // Store calculated content size for future calculations (in other functions as well)
    clientContentSize = calculateContentSize()
    return clientContentSize
  }

  function convertToWebViewPoint (point) {
    const pixelRatio = window.devicePixelRatio;
    const x = window.getComputedStyle(document.body).direction === 'rtl'
      ? (clientContentSize.width  - point.x) * pixelRatio
      : point.x * pixelRatio
    const y = point.y * pixelRatio
    return { x, y }
  }

  function getElementRect (element) {
    return BookFusion.Transformation.convertToWebViewFrameNative(element.getBoundingClientRect())
  }

  function getRangeRect (range) {
    return BookFusion.Transformation.convertToWebViewFrameNative(range.getBoundingClientRect())
  }

  function getElementPosition (element) {
    const rect = getElementRect(element)
    if (!rect) {
      return 0
    }
    return getRectPosition(rect)
  }

  function getEndElementPosition (element) {
    const rect = getElementRect(element)
    if (!rect) {
      return 0
    }
    return getEndRectPosition(rect)
  }

  function getLinkedElementPosition (linkName) {
    const element = getLinkedElement(linkName)
    return element && getElementPosition(element)
  }

    function getLinkedElementPositionById (linkName) {
      const element = getLinkedElementById(linkName)
      return element && getElementPosition(element)
    }

  function getRangePosition (range) {
    const rect = getRangeRect(range)
    if (!rect) {
      return 0
    }
    return getRectPosition(rect)
  }

  function getRectPosition (rect) {
    if (isScrollModeEnabled()) {
      return rect.y / clientContentSize.height
    } else {
      return rect.x / clientContentSize.width
    }
  }

  function getEndRectPosition (rect) {
    if (isScrollModeEnabled()) {
      return (rect.y + rect.height) / clientContentSize.height
    } else {
      return (rect.x + rect.width) / clientContentSize.width
    }
  }

  function getLinkedElement (linkName) {
    return document.querySelector(`a[name="${linkName}"]`) || document.getElementById(linkName)
  }

  function getLinkedElementById (linkName) {
    return document.getElementById(linkName)
  }

  function reportError (error) {
    BookFusionAppBridge.onErrorMessage(error)
  }

  function getPreviousElement (element) {
    if (element.previousElementSibling) {
      return element.previousElementSibling
    } else if (element.parentElement) {
      getPreviousElement(element.parentElement)
    }
  }

  function isRTL () {
    if (document.body) {
      const computedStyle = window.getComputedStyle(document.body)
      return computedStyle.direction === 'rtl' || computedStyle.writingMode === 'vertical-rl'
    }

    reportError('Empty body', EMPTY_BODY_ERROR_CODE)
    const direction = getProperty('--BF__direction')
    if (direction === 'bookfusion-direction-rtl') {
      return true
    } else if (direction === 'bookfusion-direction-ltr') {
      return false
    } else {
      return getProperty('--BF__writingMode') === 'bookfusion-writing-mode-vertical-rl'
    }
  }

  function setLanguage (language) {
     const documentElement = document.documentElement
     if (!document.documentElement) {
       reportError('Empty documentElement')
       return
     }

     if (!documentElement.getAttribute('lang')) {
       documentElement.setAttribute('lang', language)
     }
   }

  return {
    isScrollModeEnabled,
    getElementRect,
    removeProperty,
    setProperty,
    setProperties,
    getContentSize,
    getLinkedElement,
    getLinkedElementPosition,
    getLinkedElementById,
    getLinkedElementPositionById,
    getRangePosition,
    getRangeRect,
    getElementPosition,
    getEndElementPosition,
    getPreviousElement,
    convertToWebViewPoint,
    reportError,
    isRTL,
    setLanguage
  }
})()