BookFusion.Fixed = (() => {
  function parseViewportProperties (viewport) {
    return new Map(viewport
      .split(',')
      .map(keyAndValue => {
        const trimmedKeyAndValue = keyAndValue.trim()
        const [key, value] = trimmedKeyAndValue.split('=', 2)
        return [key.trim(), value.trim()]
      }))
  }

  function getViewportElement () {
    return document.querySelector('meta[name="viewport"]')
  }

  function getViewportProperties () {
    return Object.fromEntries(parseViewportProperties(getViewportElement().content))
  }

  function disableZooming () {
    const viewportElement = getViewportElement()
    if (viewportElement) {
      const viewportProperties = parseViewportProperties(viewportElement.content)
      viewportProperties.set('user-scalable', 'no')
      const newContent = Array.from(viewportProperties.entries())
        .map(([key, value]) => `${key}=${value}`)
        .join(', ')
      viewportElement.content = newContent
    } else {
      const viewportElement = document.createElement('meta')
      viewportElement.name = 'viewport'
      viewportElement.content = 'width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=no'
      const head = document.getElementsByTagName('head')[0]
      head.appendChild(viewportElement)
    }
  }

  // If any element is greater than viewport, WebKit does not properly layout it
  // e.g. book "My Brain is My Boss" page with Glossary
  function forceSize () {
    const { width, height } = getViewportProperties()
    var pixelRatio = window.devicePixelRatio;

    if (!width || !height) {
      return
    }

    document.documentElement.style.overflow = 'hidden'
    document.body.style.overflow = 'hidden'
    document.body.style.width = `${width / pixelRatio}px`
    document.body.style.height = `${height / pixelRatio}px`
  }

  return {
    getViewportProperties,
    disableZooming,
    forceSize
  }
})()