BookFusion.Font = (() => {
  let requestedFontName
  let loadedFontPaths = []

  function loadFont (name, path, { weight, style }) {
    if (loadedFontPaths.includes(path)) {
      return Promise.resolve()
    }

    const font = new FontFace(name, `url(${path})`, { weight, style })

    return font.load().then(loadedFont => {
      loadedFontPaths.push(path)
      document.fonts.add(loadedFont)
    })
  }

  function setFontFamily (name, fonts) {
    document.fonts.clear()
    loadedFontPaths = []

    requestedFontName = name

    const allFontLoaders = fonts.map(font => loadFont(name, font.path, font))

    Promise.all(allFontLoaders).then(
      () => didCompleteLoadFont(name),
      () => didFailLoadFont(name)
    )
  }

  function setDefaultFont () {
    BookFusion.Common.setProperty('--BF__fontFamily', null)

    window.BookFusionAppBridge.onFontLoaded()
  }

  function didCompleteLoadFont (name) {
    if (requestedFontName !== name) {
      return
    }

    BookFusion.Common.setProperty('--BF__fontFamily', name)

    window.BookFusionAppBridge.onFontLoaded()
  }

  function didFailLoadFont (name) {
    if (requestedFontName !== name) {
      return
    }

    window.BookFusionAppBridge.onFontFailed()
  }

  return {
    setFontFamily,
    setDefaultFont
  }
})()
