BookFusion.RangeOffsetUtils = (() => {
  function createRangeByOffsets (startOffset, endOffset) {
    const startBound = findBoundByOffset(document.body, startOffset)
    const endBound = findBoundByOffset(document.body, endOffset)

    if (startBound && endBound) {
      const range = document.createRange()
      range.setStart(...startBound)
      range.setEnd(...endBound)
      return range
    } else {
      return null
    }
  }

  function getPositionForOffset (offset) {
    const range = createRangeByOffsets(offset, offset + 1)
    if (!range) {
      return 0
    }
    return BookFusion.Common.getRangePosition(range)
  }

  function rangeIntersectsNode (range, node) {
    if (range.intersectsNode) {
      return range.intersectsNode(node)
    }

    const nodeRange = node.ownerDocument.createRange()
    try {
      nodeRange.selectNode(node)
    } catch (e) {
      nodeRange.selectNodeContents(node)
    }

    return range.compareBoundaryPoints(Range.END_TO_START, nodeRange) === -1 &&
      range.compareBoundaryPoints(Range.START_TO_END, nodeRange) === 1
  }

  function findBoundByOffset (node, offset) {
    while (node) {
      if (node.nodeType === Node.COMMENT_NODE) {
        node = node.nextSibling
      } else if (node.textContent.length <= offset) {
        offset = offset - node.textContent.length
        node = node.nextSibling
      } else if (node.childNodes.length > 0) {
        node = node.childNodes[0]
      } else {
        return [node, offset]
      }
    }
    return null
  }

  function findOffsetsByRange (range) {
    const testRange = document.createRange()
    testRange.selectNodeContents(document.body)
    testRange.setEnd(range.startContainer, range.startOffset)
    const startOffset = testRange.toString().length
    testRange.setEnd(range.endContainer, range.endOffset)
    const endOffset = testRange.toString().length

    return { startOffset, endOffset }
  }

  function getElementOffset (element) {
    const range = document.createRange()
    range.setStart(element, 0)
    range.setStart(element, 1)
    return findBoundByOffset(range)
  }

  function findOffsetsByRect (rect) {
      const { startX, startY, endX, endY } = typeof rect.startX === 'undefined'
        ? { startX: rect.x, startY: rect.y, endX: rect.x + rect.width, endY: rect.y + rect.height }
        : rect

      const startRange = document.caretRangeFromPoint(startX, startY)
      const endRange = document.caretRangeFromPoint(endX, endY)
      if (!startRange || !endRange) {
        return {}
      }

      const start = findOffsetsByRange(startRange)
      const end = findOffsetsByRange(endRange)

      const startOffset = Math.min(start.startOffset, end.endOffset)
      const endOffset = Math.max(start.startOffset, end.endOffset)

      return { startOffset, endOffset }
    }

  return {
    rangeIntersectsNode,
    getPositionForOffset,
    getElementOffset,
    findOffsetsByRange,
    createRangeByOffsets,
    findOffsetsByRect
  }
})()